// Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
// SPDX-License-Identifier: Apache-2.0

#![warn(missing_docs)]

//! The library crate that defines common helper functions that are generally used in
//! conjunction with seccompiler-bin.

pub mod backend;
pub mod common;
pub mod compiler;
/// Syscall tables
pub mod syscall_table;

use std::collections::HashMap;
use std::fmt::Debug;
use std::io::Read;
use std::sync::Arc;

use bincode::{DefaultOptions, Error as BincodeError, Options};
use common::BPF_MAX_LEN;
// Re-export the data types needed for calling the helper functions.
pub use common::{sock_filter, BpfProgram};

/// Type that associates a thread category to a BPF program.
pub type BpfThreadMap = HashMap<String, Arc<BpfProgram>>;

// BPF structure definition for filter array.
// See /usr/include/linux/filter.h .
#[repr(C)]
struct sock_fprog {
    pub len: ::std::os::raw::c_ushort,
    pub filter: *const sock_filter,
}

/// Reference to program made up of a sequence of BPF instructions.
pub type BpfProgramRef<'a> = &'a [sock_filter];

/// Binary filter deserialization errors.
#[derive(Debug, thiserror::Error, displaydoc::Display)]
pub enum DeserializationError {
    /// Bincode deserialization failed: {0}
    Bincode(BincodeError),
}

/// Filter installation errors.
#[derive(Debug, PartialEq, Eq, thiserror::Error, displaydoc::Display)]
pub enum InstallationError {
    /// Filter length exceeds the maximum size of {BPF_MAX_LEN:} instructions
    FilterTooLarge,
    /// prctl` syscall failed with error code: {0}
    Prctl(i32),
}

/// Deserialize a BPF file into a collection of usable BPF filters.
/// Has an optional `bytes_limit` that is passed to bincode to constrain the maximum amount of
/// memory that we can allocate while performing the deserialization.
/// It's recommended that the integrator of the library uses this to prevent memory allocations
/// DOS-es.
pub fn deserialize_binary<R: Read + Debug>(
    reader: R,
    bytes_limit: Option<u64>,
) -> std::result::Result<BpfThreadMap, DeserializationError> {
    let result = match bytes_limit {
        // Also add the default options. These are not part of the `DefaultOptions` as per
        // this issue: https://github.com/servo/bincode/issues/333
        Some(limit) => DefaultOptions::new()
            .with_fixint_encoding()
            .allow_trailing_bytes()
            .with_limit(limit)
            .deserialize_from::<R, HashMap<String, BpfProgram>>(reader),
        // No limit is the default.
        None => bincode::deserialize_from::<R, HashMap<String, BpfProgram>>(reader),
    };

    Ok(result
        .map_err(DeserializationError::Bincode)?
        .into_iter()
        .map(|(k, v)| (k.to_lowercase(), Arc::new(v)))
        .collect())
}

/// Helper function for installing a BPF filter.
pub fn apply_filter(bpf_filter: BpfProgramRef) -> std::result::Result<(), InstallationError> {
    // If the program is empty, don't install the filter.
    if bpf_filter.is_empty() {
        return Ok(());
    }

    // If the program length is greater than the limit allowed by the kernel,
    // fail quickly. Otherwise, `prctl` will give a more cryptic error code.
    let bpf_filter_len =
        u16::try_from(bpf_filter.len()).map_err(|_| InstallationError::FilterTooLarge)?;
    if bpf_filter_len > BPF_MAX_LEN {
        return Err(InstallationError::FilterTooLarge);
    }

    // SAFETY: Safe because the parameters are valid.
    unsafe {
        {
            let rc = libc::prctl(libc::PR_SET_NO_NEW_PRIVS, 1, 0, 0, 0);
            if rc != 0 {
                return Err(InstallationError::Prctl(*libc::__errno_location()));
            }
        }

        let bpf_prog = sock_fprog {
            len: bpf_filter_len,
            filter: bpf_filter.as_ptr(),
        };
        let bpf_prog_ptr = &bpf_prog as *const sock_fprog;
        {
            let rc = libc::prctl(
                libc::PR_SET_SECCOMP,
                libc::SECCOMP_MODE_FILTER,
                bpf_prog_ptr,
            );
            if rc != 0 {
                return Err(InstallationError::Prctl(*libc::__errno_location()));
            }
        }
    }

    Ok(())
}

#[cfg(test)]
mod tests {
    #![allow(clippy::undocumented_unsafe_blocks)]

    use std::collections::HashMap;
    use std::sync::Arc;
    use std::thread;

    use super::*;
    use crate::common::BpfProgram;

    #[test]
    fn test_deserialize_binary() {
        // Malformed bincode binary.
        {
            let data = "adassafvc".to_string();
            deserialize_binary(data.as_bytes(), None).unwrap_err();
        }

        // Test that the binary deserialization is correct, and that the thread keys
        // have been lowercased.
        {
            let bpf_prog = vec![
                sock_filter {
                    code: 32,
                    jt: 0,
                    jf: 0,
                    k: 0,
                },
                sock_filter {
                    code: 32,
                    jt: 0,
                    jf: 0,
                    k: 4,
                },
            ];
            let mut filter_map: HashMap<String, BpfProgram> = HashMap::new();
            filter_map.insert("VcpU".to_string(), bpf_prog.clone());
            let bytes = bincode::serialize(&filter_map).unwrap();

            let mut expected_res = BpfThreadMap::new();
            expected_res.insert("vcpu".to_string(), Arc::new(bpf_prog));
            assert_eq!(deserialize_binary(&bytes[..], None).unwrap(), expected_res);
        }

        // Test deserialization with binary_limit.
        {
            let bpf_prog = vec![sock_filter {
                code: 32,
                jt: 0,
                jf: 0,
                k: 0,
            }];

            let mut filter_map: HashMap<String, BpfProgram> = HashMap::new();
            filter_map.insert("t1".to_string(), bpf_prog.clone());

            let bytes = bincode::serialize(&filter_map).unwrap();

            // Binary limit too low.
            assert!(matches!(
                deserialize_binary(&bytes[..], Some(20)).unwrap_err(),
                DeserializationError::Bincode(error)
                    if error.to_string() == "the size limit has been reached"
            ));

            let mut expected_res = BpfThreadMap::new();
            expected_res.insert("t1".to_string(), Arc::new(bpf_prog));

            // Correct binary limit.
            assert_eq!(
                deserialize_binary(&bytes[..], Some(50)).unwrap(),
                expected_res
            );
        }
    }

    #[test]
    fn test_filter_apply() {
        // Test filter too large.
        thread::spawn(|| {
            let filter: BpfProgram = vec![
                sock_filter {
                    code: 6,
                    jt: 0,
                    jf: 0,
                    k: 0,
                };
                5000 // Limit is 4096
            ];

            // Apply seccomp filter.
            assert_eq!(
                apply_filter(&filter).unwrap_err(),
                InstallationError::FilterTooLarge
            );
        })
        .join()
        .unwrap();

        // Test empty filter.
        thread::spawn(|| {
            let filter: BpfProgram = vec![];

            assert_eq!(filter.len(), 0);

            let seccomp_level = unsafe { libc::prctl(libc::PR_GET_SECCOMP) };
            assert_eq!(seccomp_level, 0);

            apply_filter(&filter).unwrap();

            // test that seccomp level remains 0 on failure.
            let seccomp_level = unsafe { libc::prctl(libc::PR_GET_SECCOMP) };
            assert_eq!(seccomp_level, 0);
        })
        .join()
        .unwrap();

        // Test invalid BPF code.
        thread::spawn(|| {
            let filter = vec![sock_filter {
                // invalid opcode
                code: 9999,
                jt: 0,
                jf: 0,
                k: 0,
            }];

            let seccomp_level = unsafe { libc::prctl(libc::PR_GET_SECCOMP) };
            assert_eq!(seccomp_level, 0);

            assert_eq!(
                apply_filter(&filter).unwrap_err(),
                InstallationError::Prctl(22)
            );

            // test that seccomp level remains 0 on failure.
            let seccomp_level = unsafe { libc::prctl(libc::PR_GET_SECCOMP) };
            assert_eq!(seccomp_level, 0);
        })
        .join()
        .unwrap();
    }
}
